#include<FastLED.h>

#define LED_PIN     6
#define BRIGHTNESS  60
#define LED_TYPE    WS2812B
#define COLOR_ORDER GRB

const uint8_t kMatrixWidth  = 16;
const uint8_t kMatrixHeight = 16;
const bool    kMatrixSerpentineLayout = true;


#define NUM_LEDS (kMatrixWidth * kMatrixHeight)
#define MAX_DIMENSION ((kMatrixWidth>kMatrixHeight) ? kMatrixWidth : kMatrixHeight)

// The leds
CRGB leds[kMatrixWidth * kMatrixHeight];

// The 16 bit version of our coordinates
static uint16_t x;
static uint16_t y;
static uint16_t z;

uint16_t speed = 10; // speed is set dynamically once we've started up

uint16_t scale = 25; // scale is set dynamically once we've started up

uint8_t noise[MAX_DIMENSION][MAX_DIMENSION];

CRGBPalette16 currentPalette( CloudColors_p );
uint8_t       colorLoop = 1;

void setup() {
  delay(3000);
  LEDS.addLeds<LED_TYPE,LED_PIN,COLOR_ORDER>(leds,NUM_LEDS);
  LEDS.setBrightness(BRIGHTNESS);

  // Initialize our coordinates to some random values
  x = random16();
  y = random16();
  z = random16();
}

void fillnoise8() {
  uint8_t dataSmoothing = 0;
  if( speed < 50) {
    dataSmoothing = 400 - (speed * 4);
  }
  
  for(int i = 0; i < MAX_DIMENSION; i++) {
    int ioffset = scale * i;
    for(int j = 0; j < MAX_DIMENSION; j++) {
      int joffset = scale * j;
      
      uint8_t data = inoise8(x + ioffset,y + joffset,z);
      data = qsub8(data,16);
      data = qadd8(data,scale8(data,39));

      if( dataSmoothing ) {
        uint8_t olddata = noise[i][j];
        uint8_t newdata = scale8( olddata, dataSmoothing) + scale8( data, 256 - dataSmoothing);
        data = newdata;
      }
      
      noise[i][j] = data;
    }
  }
  
  z += speed;

  x += speed / 8;
  y -= speed / 16;
}

void mapNoiseToLEDsUsingPalette()
{
  static uint8_t ihue=0;
  
  for(int i = 0; i < kMatrixWidth; i++) {
    for(int j = 0; j < kMatrixHeight; j++) {

      uint8_t index = noise[j][i];
      uint8_t bri =   noise[i][j];

      if( colorLoop) { 
        index += ihue;
      }

      if( bri > 127 ) {
        bri = 255;
      } else {
        bri = dim8_raw( bri * 2);
      }

      CRGB color = ColorFromPalette( currentPalette, index, bri);
      leds[XY(i,j)] = color;
    }
  }
  
  ihue+=1;
}

void loop() {

  ChangePaletteAndSettingsPeriodically();

   fillnoise8();
  
  mapNoiseToLEDsUsingPalette();

  LEDS.show();
}


#define HOLD_PALETTES_X_TIMES_AS_LONG 10

void ChangePaletteAndSettingsPeriodically()
{
  uint8_t secondHand = ((millis() / 1000) / HOLD_PALETTES_X_TIMES_AS_LONG) % 60;
  static uint8_t lastSecond = 99;
  
  if( lastSecond != secondHand) {
    lastSecond = secondHand;
//if( secondHand ==  0)  { currentPalette = CloudColors_p;           speed =  1; scale = 30; colorLoop = 1; }
    if( secondHand ==  0)  { SetupRandomPalette();                     speed =  1; scale = 30; colorLoop = 1; }
    if( secondHand ==  5)  { SetupRandomPalette();                     speed =  1; scale = 30; colorLoop = 1; }
//    if( secondHand == 10)  { currentPalette = ForestColors_p;          speed =  1; scale = 30; colorLoop = 0; }
    if( secondHand == 10)  { SetupRandomPalette();                     speed =  1; scale = 30; colorLoop = 1; }
//    if( secondHand == 15)  { currentPalette = CloudColors_p;           speed =  1; scale = 30; colorLoop = 0; }  
    if( secondHand == 15)  { SetupRandomPalette();                     speed =  1; scale = 30; colorLoop = 1; }  
//    if( secondHand == 20)  { currentPalette = OceanColors_p;           speed =  1; scale = 30; colorLoop = 0; }
    if( secondHand == 20)  { SetupRandomPalette();                     speed =  1; scale = 30; colorLoop = 1; }
    if( secondHand == 25)  { SetupRandomPalette();                     speed =  1; scale = 30; colorLoop = 0; }
    if( secondHand == 30)  { SetupRandomPalette();                     speed =  1; scale = 30; colorLoop = 1; }
    if( secondHand == 35)  { SetupRandomPalette();                     speed =  1; scale = 30; colorLoop = 1; }
    if( secondHand == 40)  { SetupRandomPalette();                     speed =  1; scale = 30; colorLoop = 1; }
    if( secondHand == 45)  { SetupRandomPalette();                     speed =  1; scale = 30; colorLoop = 1; }
    if( secondHand == 50)  { SetupRandomPalette();                     speed =  1; scale = 30; colorLoop = 1; }
 //   if( secondHand == 50)  { currentPalette = RainbowStripeColors_p;   speed =  1; scale = 30; colorLoop = 1; }
  }
}

void SetupRandomPalette()
{
  currentPalette = CRGBPalette16( 
                      CHSV( random8(), 255, 32), 
                      CHSV( random8(), 255, 255), 
                      CHSV( random8(), 128, 255), 
                      CHSV( random8(), 255, 255)); 
}

void SetupBlackAndWhiteStripedPalette()
{
  fill_solid( currentPalette, 16, CRGB::Black);
  currentPalette[0] = CRGB::White;
  currentPalette[4] = CRGB::White;
  currentPalette[8] = CRGB::White;
  currentPalette[12] = CRGB::White;
}

uint16_t XY( uint8_t x, uint8_t y)
{
  uint16_t i;
  if( kMatrixSerpentineLayout == false) {
    i = (y * kMatrixWidth) + x;
  }
  if( kMatrixSerpentineLayout == true) {
    if( y & 0x01) {
      // Odd rows run backwards
      uint8_t reverseX = (kMatrixWidth - 1) - x;
      i = (y * kMatrixWidth) + reverseX;
    } else {
      // Even rows run forwards
      i = (y * kMatrixWidth) + x;
    }
  }
  return i;
}
